# Balance Power-Performance parent class.

import math


def initChooseMachine():
    """
    Initialization of variables for the functions chooseMachineAndState and chooseMachineAndStateBase.
    """
    chosenMachine = -1 # Variable that determines the type of machine to be switched on.
    indexChosenMachine = -1 # In the list ratioAndComputingPowerMachine, index of the chosen machine.
    maximumScore = 0 # Variable saving the maximum score reached.
    return chosenMachine, indexChosenMachine, maximumScore


def computeMinRatioAndMaxComputingPower(ratioAndComputingPowerMachine):
    """
    Function that computes the smallest ratio and the largest computing power in the ratioAndPowerComputingMachine
    list of possibles machines and DVFS states combinations. This function is used in chooseMachineAndState
    and chooseMachineAndStateBase.

    :param ratioAndComputingPowerMachine: A list containing the machines that can be switched on with their
    information ([type of machine, DVFS state, ratio, computing power, requested power]) and their associated score.
    """
    maxComputingPower = 0
    minRatio = math.inf
    for computing in ratioAndComputingPowerMachine:
        if computing[3] > maxComputingPower:
            maxComputingPower = computing[3]
        if computing[2] < minRatio:
            minRatio = computing[2]
    return maxComputingPower, minRatio


class BalancePowerPerformance:
    def __init__(self, listMachine, expe, resultHeuristicExists):
        """
        The BalancePowerPerformance parent class.

        :param listMachine: List containing the machine objects (number, static power, states and name).
        :param expe: Experience object (powerStart, powerEnd et byStep).
        :param resultHeuristicExists: Boolean. True if the result files of this experiment with these machines
        already exist, False if not.
        """
        self.listMachine = listMachine
        self.expe = expe
        self.resultHeuristicExists = resultHeuristicExists
        self.objective = []
        self.runtime = -1

    def init(self):
        """
        Initialization of variables. This function is used in BPPHeuristic function.
        """
        stop = False
        objectiveWithCurrentPower = 0
        numberMachineAvailable = []
        for k in range(len(self.listMachine)):
            numberMachineAvailable.append(self.listMachine[k].number)
        powerMax = 0 # Variable indicating the maximum power that can be requested by a machine.
        for k in range(len(self.listMachine)):
            currentPowerMax = self.listMachine[k].static + self.listMachine[k].states[-1][0] * \
                              self.listMachine[k].states[-1][1]
            if currentPowerMax > powerMax:
                powerMax = currentPowerMax
        return stop, objectiveWithCurrentPower, numberMachineAvailable, powerMax

    def computeRatiosAndComputingPowerBase(self, numberMachineAvailable):
        """
        If the available power is higher than the maximum power that can be requested by the machines, we compute the
        ratio, the computing power and the requested power for each type of machine and for each DVFS state.

        :param numberMachineAvailable: Number of machines available for each type of machine.
        """
        ratioAndComputingPowerMachineBase = []
        for k in range(len(self.listMachine)):
            if numberMachineAvailable[k] > 0:
                for i in range(1, len(self.listMachine[k].states)):
                    dynamicPowerMachine = self.listMachine[k].states[i][0] * self.listMachine[k].states[i][1]
                    computingPowerMachine = dynamicPowerMachine / self.listMachine[k].states[i][1]
                    ratioAndComputingPowerMachineBase.append([k, i, (self.listMachine[k].static + dynamicPowerMachine) /
                                                              computingPowerMachine, computingPowerMachine,
                                                              self.listMachine[k].static + dynamicPowerMachine])
        return ratioAndComputingPowerMachineBase

    def computeRatiosAndComputingPower(self, power, numberMachineAvailable):
        """
        If the available power is lower than the maximum power that can be requested by the machines, the ratio,
        the computing power and the requested power for each type of machine and for each DVFS state are computed
        according to the available power.

        :param power: The available power.
        :param numberMachineAvailable: Number of available machines for each type of machine.
        """
        ratioAndComputingPowerMachine = []
        for k in range(len(self.listMachine)):
            if numberMachineAvailable[k] > 0:
                if power > self.listMachine[k].static:
                    for i in range(1, len(self.listMachine[k].states)):
                        power_p = power - self.listMachine[k].static
                        dynamicPowerMachine = min(self.listMachine[k].states[i][0] * self.listMachine[k].states[i][1],
                                                  power_p)
                        computingPowerMachine = dynamicPowerMachine / self.listMachine[k].states[i][1]
                        ratioAndComputingPowerMachine.append([k, i, (self.listMachine[k].static + dynamicPowerMachine) /
                                                              computingPowerMachine, computingPowerMachine,
                                                              self.listMachine[k].static + dynamicPowerMachine])
        return ratioAndComputingPowerMachine

    @staticmethod
    def chooseMachineAndStateBase(ratioAndComputingPowerMachine, alpha):
        """
        Function that determines which machine to switch on (the one with the highest score) when the available power 
        is higher than the maximum power that can be requested by the machines.

        :param ratioAndComputingPowerMachine: A list containing the machines that can be switched on with their
        information ([type of machine, DVFS state, ratio, computing power, requested power]) and their associated score.
        :param alpha: Real value between 0 and 1, depending on what you want to maximize in the computation of the
        machine scores: The computing power (1) or the ratio (0).
        """
        chooseMachine, indexChoiceMachine, maxScore = initChooseMachine()
        maxComputingPower, minRatio = computeMinRatioAndMaxComputingPower(ratioAndComputingPowerMachine)
        for i in range(len(ratioAndComputingPowerMachine)):
            score = alpha * ratioAndComputingPowerMachine[i][3] / maxComputingPower + (1 - alpha) * minRatio / \
                    ratioAndComputingPowerMachine[i][2]
            if score > maxScore:
                chooseMachine = ratioAndComputingPowerMachine[i][0]
                indexChoiceMachine = i
                maxScore = score
        objective = ratioAndComputingPowerMachine[indexChoiceMachine][3]
        powerToBeSubstracted = ratioAndComputingPowerMachine[indexChoiceMachine][4]
        machineToBeSubstracted = chooseMachine
        return powerToBeSubstracted, machineToBeSubstracted, objective

    @staticmethod
    def chooseMachineAndState(power, objectiveWithCurrentPower, numberMachineAvailable,
                              ratioAndComputingPowerMachine, stop, alpha):
        """
        Function that determines which machine to switch on (the one with the highest score) in the case where the
        available power is lower than the maximum power that can be requested by the machines.
        """
        if not ratioAndComputingPowerMachine:
            stop = True
        else:
            chooseMachine, indexChoiceMachine, maxScore = initChooseMachine()
            maxComputingPower, minRatio = computeMinRatioAndMaxComputingPower(ratioAndComputingPowerMachine)
            for i in range(len(ratioAndComputingPowerMachine)):
                score = alpha * ratioAndComputingPowerMachine[i][3] / maxComputingPower + (1 - alpha) * minRatio / \
                        ratioAndComputingPowerMachine[i][2]
                if score > maxScore and numberMachineAvailable[ratioAndComputingPowerMachine[i][0]] > 0:
                    chooseMachine = ratioAndComputingPowerMachine[i][0]
                    indexChoiceMachine = i
                    maxScore = score
            objectiveWithCurrentPower += ratioAndComputingPowerMachine[indexChoiceMachine][3]
            power -= ratioAndComputingPowerMachine[indexChoiceMachine][4]
            numberMachineAvailable[chooseMachine] -= 1
        return power, stop, numberMachineAvailable, objectiveWithCurrentPower

    def BPPHeuristic(self, power, alpha):
        """
        Function that computes the computing power feasible by the machines, depending on the given power and the
        alpha parameter.
        """
        stop, objectiveWithCurrentPower, numberMachineAvailable, powerMax = self.init()
        while not stop:
            if sum(numberMachineAvailable) == 0:
                break
            if power < powerMax:
                ratioAndComputingPowerMachine = self.computeRatiosAndComputingPower(power, numberMachineAvailable)
                power, stop, numberMachineAvailable, objectiveWithCurrentPower = \
                    self.chooseMachineAndState(power, objectiveWithCurrentPower, numberMachineAvailable,
                                               ratioAndComputingPowerMachine, stop, alpha)
            else:
                ratioAndComputingPowerMachine = self.computeRatiosAndComputingPowerBase(numberMachineAvailable)
                powerToBeSubstracted, machineToBeSubstracted, objective = \
                    self.chooseMachineAndStateBase(ratioAndComputingPowerMachine, alpha)
                while power > powerMax and numberMachineAvailable[machineToBeSubstracted] > 0:
                    power -= powerToBeSubstracted
                    numberMachineAvailable[machineToBeSubstracted] -= 1
                    objectiveWithCurrentPower += objective
        return objectiveWithCurrentPower
