# BSRWS subclass

from bestStateClass import BestStateClass, init
import time


class BSRWS(BestStateClass):
    def __init__(self, listMachine, expe, resultHeuristicExists):
        """
        The BSRWS subclass.

        :param listMachine: List containing the machine objects (number, static power, states and name).
        :param expe: Experience object (powerStart, powerEnd et byStep).
        :param resultHeuristicExists: Boolean. True if the result files of this experiment with these machines
        already exist, False if not.
        """
        super().__init__(listMachine, expe, resultHeuristicExists)
        self.xRuntime = []

    def getObjective(self):
        return self.objective

    def getRuntime(self):
        return self.runtime

    def getXRuntime(self):
        return self.xRuntime

    def chooseMachineToSwitchOn(self, power, powerMax, numberMachineAvailable, chosenMachineBase, chosenStateBase,
                                possibleToSwitchOn, solution, solutionNumber):
        """
        Function that determines the machine to be switched on.

        :param power: The available power.
        :param powerMax: The maximum power that can be requested (of the most consuming machine).
        :param numberMachineAvailable: Number of machines available for each type of machine.
        :param chosenMachineBase: The chosen machine in the case where the available power is higher than the
        maximum power that can be requested and if this machine is available.
        :param chosenStateBase: The selected DVFS state of the chosen machine in the case where the available power
        is higher than the maximum power that can be requested and if this machine is available.
        :param possibleToSwitchOn: Boolean variable indicating if it is possible to switch on a machine.
        :param solution: The list containing the lists of switched on machines and their information.
        :param solutionNumber: The list containing the number of machines switched on.
        """
        if power > powerMax:
            if numberMachineAvailable[chosenMachineBase] > 0:
                chosenMachine = chosenMachineBase
                chosenState = chosenStateBase
            else:
                ratioMachineBase = self.computeRatiosBase(numberMachineAvailable)
                chosenMachine, chosenState, possibleToSwitchOn = \
                    self.chooseMachineAndState(ratioMachineBase, possibleToSwitchOn)
                chosenMachineBase = chosenMachine
                chosenStateBase = chosenState
        else:
            ratioMachine = self.computeRatios(power, numberMachineAvailable)
            chosenMachine, chosenState, possibleToSwitchOn = \
                self.chooseMachineAndState(ratioMachine, possibleToSwitchOn)
            if chosenMachine == -1:
                possibleToSwitchOn = False
        if possibleToSwitchOn:
            power, numberMachineAvailable, solution, solutionNumber = \
                self.switchOnMachine(power, chosenMachine, chosenState,
                                     numberMachineAvailable, solution, solutionNumber)
        return power, numberMachineAvailable, solution, solutionNumber, possibleToSwitchOn, chosenMachineBase, \
               chosenStateBase

    def bestStateRedistributeWithoutStatic(self, power, numberMachineAvailable):
        """
        The BSRWS heuristic.

        :param power: The available power.
        :param numberMachineAvailable: Number of machines available for each type of machine.
        """
        powerMax = 0
        for k in range(len(self.listMachine)):
            if self.listMachine[k].static + self.listMachine[k].states[-1][0] * self.listMachine[k].states[-1][1] > \
                    powerMax:
                powerMax = self.listMachine[k].static + self.listMachine[k].states[-1][0] * \
                           self.listMachine[k].states[-1][1]
        possibleToSwitchOn, solution, solutionNumber = init()
        chosenMachineBase = -1
        chosenStateBase = -1
        if power > powerMax:
            ratioMachineBase = self.computeRatiosBase(numberMachineAvailable)
            chosenMachineBase, chosenStateBase, possibleToSwitchOn = \
                self.chooseMachineAndState(ratioMachineBase, possibleToSwitchOn)
        while possibleToSwitchOn:
            power, numberMachineAvailable, solution, solutionNumber, possibleToSwitchOn, chosenMachineBase, \
            chosenStateBase = self.chooseMachineToSwitchOn(power, powerMax, numberMachineAvailable, chosenMachineBase,
                                                           chosenStateBase, possibleToSwitchOn, solution,
                                                           solutionNumber)
        if not solution and power >= 0:
            return [], []
        elif solution and power >= 0:
            stop = False
            while not stop:
                solution, solutionNumber, stop, power = \
                    self.powerAllocationWithoutStatic(power, stop, solution, solutionNumber)
        return solution, solutionNumber

    def computeBSRWS(self):
        """
        Function that computes the BSRWS heuristic.
        """
        if not self.resultHeuristicExists:
            start = time.time()
            power = self.expe.powerStart
            start2 = time.time()
            numberMachineAvailable = self.getMachineMax()
            solution, solutionNumber = self.bestStateRedistributeWithoutStatic(power, numberMachineAvailable)
            currentObjectif = self.computeObjectiveSolution(solution, solutionNumber)
            end2 = time.time()
            self.xRuntime.append(end2 - start2)
            self.objective.append(currentObjectif)
            end = time.time()
            self.runtime = end - start
