# BSRWSAR subclass

from bestStateClass import BestStateClass, init
import time


class BSRWSAndRemoving(BestStateClass):
    def __init__(self, listMachine, expe, resultHeuristicExists):
        """
        The BSRWSAR subclass.

        :param listMachine: List containing the machine objects (number, static power, states and name).
        :param expe: Experience object (powerStart, powerEnd et byStep).
        :param resultHeuristicExists: Boolean. True if the result files of this experiment with these machines
        already exist, False if not.
        """
        super().__init__(listMachine, expe, resultHeuristicExists)
        self.xRuntime = []

    def getObjective(self):
        return self.objective

    def getRuntime(self):
        return self.runtime

    def getXRuntime(self):
        return self.xRuntime

    def computePowerMax(self):
        """
        Function that computes the maximum power that can be requested by a machine.
        """
        powerMax = 0
        for k in range(len(self.listMachine)):
            if self.listMachine[k].static + self.listMachine[k].states[-1][0] * self.listMachine[k].states[-1][1] > \
                    powerMax:
                powerMax = self.listMachine[k].static + self.listMachine[k].states[-1][0] * \
                           self.listMachine[k].states[-1][1]
        return powerMax

    def computeMachineToSwitchOn(self, power, powerMax, numberMachineAvailable, chosenMachineBase,
                                 chosenStateBase, possibleToSwitchOn, solution, solutionNumber):
        """
        Function that determines the machine to be switched on.

        :param power: The available power.
        :param powerMax: The maximum power that can be requested (of the most consuming machine).
        :param numberMachineAvailable: Number of machines available for each type of machine.
        :param chosenMachineBase: The chosen machine in the case where the available power is higher than the
        maximum power that can be requested and if this machine is available.
        :param chosenStateBase: The selected DVFS state of the chosen machine in the case where the available power
        is higher than the maximum power that can be requested and if this machine is available.
        :param possibleToSwitchOn: Boolean variable indicating if it is possible to switch on a machine.
        :param solution: The list containing the lists of switched on machines and their information.
        :param solutionNumber: The list containing the number of machines switched on.
        """
        if power > powerMax:
            if numberMachineAvailable[chosenMachineBase] > 0:
                chosenMachine = chosenMachineBase
                chosenState = chosenStateBase
            else:
                ratioMachineBase = self.computeRatiosBase(numberMachineAvailable)
                chosenMachine, chosenState, possibleToSwitchOn = \
                    self.chooseMachineAndState(ratioMachineBase, possibleToSwitchOn)
                chosenMachineBase = chosenMachine
                chosenStateBase = chosenState
        else:
            ratioMachine = self.computeRatios(power, numberMachineAvailable)
            chosenMachine, chosenState, possibleToSwitchOn = \
                self.chooseMachineAndState(ratioMachine, possibleToSwitchOn)
            if chosenMachine == -1:
                possibleToSwitchOn = False
        if possibleToSwitchOn:
            power, numberMachineAvailable, solution, solutionNumber = \
                self.switchOnMachine(power, chosenMachine, chosenState,
                                     numberMachineAvailable, solution, solutionNumber)
        return solution, solutionNumber, power, chosenMachineBase, chosenStateBase, possibleToSwitchOn

    def BSRWS(self, power, powerMax, numberMachineAvailable, possibleToSwitchOn, chosenMachineBase, chosenStateBase,
              solution, solutionNumber, removing, permanentSolution, permanentSolutionNumber, permanentPower):
        """
        Function that computes the first solution (the machines to turn on according to the power) and the
        permanent solution if the boolean variable removing is true.

        :param power: The available power.
        :param powerMax: The maximum power that can be requested (of the most consuming machine).
        :param numberMachineAvailable: Number of machines available for each type of machine.
        :param chosenMachineBase: The chosen machine in the case where the available power is higher than the
        maximum power that can be requested and if this machine is available.
        :param chosenStateBase: The selected DVFS state of the chosen machine in the case where the available power
        is higher than the maximum power that can be requested and if this machine is available.
        :param possibleToSwitchOn: Boolean variable indicating if it is possible to switch on a machine.
        :param solution: The list containing the lists of switched on machines and their information.
        :param solutionNumber: The list containing the number of machines switched on.
        :param removing: Boolean variable recording permanentSolution and permanentSolutionNumber if True.
        :param permanentSolution: Permanent list solution.
        :param permanentSolutionNumber: Permanent list solutionNumber.
        :param permanentPower: Permanent power value.
        """
        if power > powerMax:
            ratioMachineBase = self.computeRatiosBase(numberMachineAvailable)
            chosenMachineBase, chosenStateBase, possibleToSwitchOn = \
                self.chooseMachineAndState(ratioMachineBase, possibleToSwitchOn)
        while possibleToSwitchOn:
            solution, solutionNumber, power, chosenMachineBase, chosenStateBase, possibleToSwitchOn = \
                self.computeMachineToSwitchOn(power, powerMax, numberMachineAvailable, chosenMachineBase,
                                              chosenStateBase, possibleToSwitchOn, solution, solutionNumber)
        if not solution:
            return [], [], [], [], 0
        else:
            if removing:
                permanentSolution = solution[:]
                permanentSolutionNumber = solutionNumber[:]
                permanentPower = power
            stop = False
            while not stop:
                solution, solutionNumber, stop, power = \
                    self.powerAllocationWithoutStatic(power, stop, solution, solutionNumber)
        return solution, solutionNumber, permanentSolution, permanentSolutionNumber, permanentPower

    def removing(self, permanentSolution, permanentSolutionNumber, permanentPower):
        """
        Function that takes over the permanent solution by removing the last machine switched on and allocates
        the available power to the remaining switched on machines.

        :param permanentSolution: Permanent list solution.
        :param permanentSolutionNumber: Permanent list solutionNumber.
        :param permanentPower: Permanent power value.
        """
        if permanentSolutionNumber[-1] > 1:
            k = permanentSolution[-1][0]
            permanentPower += self.listMachine[k].static + permanentSolution[-1][2]
            permanentSolutionNumber[-1] -= 1
        else:
            k = permanentSolution[-1][0]
            permanentPower += self.listMachine[k].static + permanentSolution[-1][2]
            del permanentSolutionNumber[-1]
            del permanentSolution[-1]
        stop = False
        solution = permanentSolution[:]
        solutionNumber = permanentSolutionNumber[:]
        power = permanentPower
        while not stop:
            solution, solutionNumber, stop, power = \
                self.powerAllocationWithoutStatic(power, stop, solution, solutionNumber)
        return solution, solutionNumber, permanentSolution, permanentSolutionNumber, permanentPower

    def bestStateRedistributeWithoutStaticAndRemoving(self, power, removing, permanentSolution,
                                                      numberMachineAvailable, permanentSolutionNumber, permanentPower):
        """
        The BSRWSAR heuristic.

        :param power: The available power.
        :param numberMachineAvailable: Number of machines available for each type of machine.
        :param removing: Boolean variable recording permanentSolution and permanentSolutionNumber if True.
        :param permanentSolution: Permanent list solution.
        :param permanentSolutionNumber: Permanent list solutionNumber.
        :param permanentPower: Permanent power value.
        """
        possibleToSwitchOn, solution, solutionNumber = init()
        chosenMachineBase = -1
        chosenStateBase = -1
        powerMax = self.computePowerMax()
        if not permanentSolution:
            solution, solutionNumber, permanentSolution, permanentSolutionNumber, permanentPower = \
                self.BSRWS(power, powerMax, numberMachineAvailable, possibleToSwitchOn, chosenMachineBase,
                           chosenStateBase, solution, solutionNumber, removing, permanentSolution,
                           permanentSolutionNumber, permanentPower)
        else:
            solution, solutionNumber, permanentSolution, permanentSolutionNumber, permanentPower = \
                self.removing(permanentSolution, permanentSolutionNumber, permanentPower)
        return solution, solutionNumber, permanentSolution, permanentSolutionNumber, permanentPower

    def computeBSRWSAR(self):
        """
        Function that computes the BSRWSAR heuristic.
        """
        removing = True
        if not self.resultHeuristicExists:
            start = time.time()
            for power in range(self.expe.powerStart, self.expe.powerEnd, self.expe.byStep):
                stop = False
                objectif = 0
                permanentSolution = []
                permanentSolutionNumber = []
                permanentPower = 0
                start2 = time.time()
                numberMachineAvailable = self.getMachineMax()
                saveNumberMach = True
                numberMach = 0
                while not stop:
                    solution, solutionNumber, permanentSolution, permanentSolutionNumber, permanentPower = \
                        self.bestStateRedistributeWithoutStaticAndRemoving(power, removing, permanentSolution,
                                                                           numberMachineAvailable,
                                                                           permanentSolutionNumber, permanentPower)
                    currentObjectif = self.computeObjectiveSolution(solution, solutionNumber)
                    if saveNumberMach:
                        numberMach = sum(solutionNumber)
                        saveNumberMach = False
                    if currentObjectif >= objectif:
                        objectif = currentObjectif
                    if sum(solutionNumber) < 0.01 * numberMach:
                        self.objective.append(objectif)
                        stop = True
                end2 = time.time()
                self.xRuntime.append(end2 - start2)
            end = time.time()
            self.runtime = end - start
